"""
Classes and functions to treat column data types according to
Deirokay data types.
"""

from typing import Iterable, List, Optional, Union

import dask.dataframe  # lazy module
import numpy  # lazy module
import pandas  # lazy module

from deirokay._typing import DeirokaySerializedSeries
from deirokay.enums import Backend, DTypes

from ..multibackend import serialize, treat
from .validator import Validator


class BooleanTreater(Validator):
    """Treater for boolean-like variables

    Parameters
    ----------
    truthies : List[str], optional
        List of strings to be considered as True,
        by default ['true', 'True']
    falsies : List[str], optional
        List of strings to be considered as False,
        by default ['false', 'False']
    ignore_case : bool, optional
        Ignore case when evaluating list of values,
        by default False
    default_value : Optional[bool], optional
        Value to replace null values (None, pandas.NA or numpy,NaN).
        By default None.
    """
    supported_backends = [Backend.PANDAS, Backend.DASK]
    supported_dtype = DTypes.BOOLEAN
    supported_primitives = [bool]

    def __init__(self,
                 truthies: List[str] = ['true', 'True'],
                 falsies: List[str] = ['false', 'False'],
                 ignore_case: bool = False,
                 default_value: Optional[bool] = None,
                 **kwargs):
        super().__init__(**kwargs)

        assert default_value in (True, False, None)

        self.ignore_case = ignore_case
        self.default_value = default_value

        if self.ignore_case:
            self.truthies = set(truthy.lower() for truthy in truthies)
            self.falsies = set(falsy.lower() for falsy in falsies)
        else:
            self.truthies = set(truthies)
            self.falsies = set(falsies)

        if self.truthies & self.falsies:
            raise ValueError('Truthies and Falsies sets should be'
                             ' disjoint sets.')

    def _evaluate(self, value: Union[bool, str, None]) -> Union[bool, None]:
        if value is True:
            return True
        if value is False:
            return False
        if value is None or value is numpy.nan or value is pandas.NA:
            return self.default_value

        _value = value.lower() if self.ignore_case else value
        if _value in self.truthies:
            return True
        if _value in self.falsies:
            return False
        raise ValueError(f'Unexpected boolean value: "{value}"'
                         f' ({value.__class__})\n'
                         f'Expected values: {self.truthies | self.falsies}')

    @treat(Backend.PANDAS)
    def _treat_pandas(self, series: Iterable) -> 'pandas.Series':
        series = super()._treat_pandas(series)
        series = series.apply(self._evaluate).astype('boolean')
        # Validate again
        super()._treat_pandas(series)

        return series

    @treat(Backend.DASK)
    def _treat_dask(
        self, series: Iterable
    ) -> 'dask.dataframe.Series':
        series = super()._treat_dask(series)
        series = series.apply(
            self._evaluate, meta=(series.name, 'object')
        ).astype('boolean')
        # Validate again
        super()._treat_dask(series)

        return series

    @staticmethod
    def _serialize_common(series):
        def _convert(item):
            if item is pandas.NA:
                return None
            return bool(item)
        return {
            'values': [_convert(item) for item in series],
            'parser': {
                'dtype': BooleanTreater.supported_dtype.value
            }
        }

    @serialize(Backend.PANDAS)
    @staticmethod
    def _serialize_pandas(
        series: 'pandas.Series'
    ) -> DeirokaySerializedSeries:
        return BooleanTreater._serialize_common(series)

    @serialize(Backend.DASK)
    @staticmethod
    def _serialize_dask(
        series: 'dask.dataframe.Series'
    ) -> DeirokaySerializedSeries:
        return BooleanTreater._serialize_common(series)



"""
Classes and functions to treat column data types according to
Deirokay data types.
"""
from typing import Iterable

import dask.dataframe  # lazy module
import pandas  # lazy module

from deirokay._typing import DeirokaySerializedSeries
from deirokay.enums import Backend, DTypes

from ..multibackend import serialize, treat
from .numeric_treater import NumericTreater


class IntegerTreater(NumericTreater):
    """Treater for integer variables"""
    supported_backends = [Backend.PANDAS, Backend.DASK]
    supported_dtype = DTypes.INT64

    @treat(Backend.PANDAS)
    def _treat_pandas(self, series: Iterable) -> 'pandas.Series':
        return super()._treat_pandas(series).astype(float).astype('Int64')

    @treat(Backend.DASK)
    def _treat_dask(
        self, series: Iterable
    ) -> 'dask.dataframe.Series':
        return super()._treat_dask(series).astype(float).astype('Int64')

    @staticmethod
    def _serialize_common(series):
        def _convert(item):
            if item is pandas.NA:
                return None
            return int(item)
        return {
            'values': [_convert(item) for item in series],
            'parser': {
                'dtype': IntegerTreater.supported_dtype.value
            }
        }

    @serialize(Backend.PANDAS)
    @staticmethod
    def _serialize_pandas(
        series: 'pandas.Series'
    ) -> DeirokaySerializedSeries:
        return IntegerTreater._serialize_common(series)

    @serialize(Backend.DASK)
    @staticmethod
    def _serialize_dask(
        series: 'dask.dataframe.Series'
    ) -> DeirokaySerializedSeries:
        return IntegerTreater._serialize_common(series)



"""
Deirokay's main set of classes and functions.
"""

from typing import TYPE_CHECKING

if not TYPE_CHECKING:
    import lazy_import
    lazy_import.lazy_module('dask.dataframe')
    lazy_import.lazy_module('numpy')
    lazy_import.lazy_module('pandas')

from .__version__ import __version__  # noqa: F401
from .parser import data_reader
from .profiler import profile
from .validator import validate

__all__ = (
    'data_reader',
    'profile',
    'validate',
)



import functools
from types import ModuleType
from typing import Type

from deirokay._typing import DeirokayStatement
from deirokay.enums import Backend
from deirokay.fs import fs_factory
from deirokay.statements import BaseStatement

from . import STATEMENTS_MAP


@functools.lru_cache(maxsize=None)
def _cached_import_file_as_module(file_path: str) -> ModuleType:
    """Import a file as a module, caching the result.
    This prevents the need to read the same file multiple times.
    """
    fs = fs_factory(file_path)
    module = fs.import_as_python_module()
    return module


def _load_custom_statement(location: str) -> Type[BaseStatement]:
    """Load a custom statement from a .py file"""
    if '::' not in location:
        raise ValueError('You should pass your class location using the'
                         ' following pattern:\n'
                         '<.py file location>::<class name>')

    file_path, class_name = location.split('::')

    module = _cached_import_file_as_module(file_path)
    cls = getattr(module, class_name)

    if not issubclass(cls, BaseStatement):
        raise ImportError('Your custom statement should be a subclass of'
                          ' BaseStatement')

    return cls


def statement_factory(statement: DeirokayStatement,
                      backend: Backend) -> BaseStatement:
    """Receive statement dict and create the proper statement object.
    The `name` attribute of the class is used to bind the statement
    type to its class.

    Parameters
    ----------
    statement : dict
        Dict from Validation Document representing statement and its
        parameters.

    Returns
    -------
    BaseStatement
        Instance of Statement class.

    Raises
    ------
    KeyError
        Custom statement should present a `location` default parameter.
    NotImplementedError
        Declared statement type does not exist.
    """
    stmt_type = statement.get('type')

    if stmt_type == 'custom':
        location = statement.get('location')
        if not location:
            raise KeyError('A custom statement must define a `location`'
                           ' parameter.')
        cls = _load_custom_statement(location)

    elif stmt_type in STATEMENTS_MAP:
        cls = STATEMENTS_MAP[stmt_type]

    else:
        raise NotImplementedError(
            f'Statement type "{stmt_type}" not implemented.\n'
            f'The available types are {list(STATEMENTS_MAP)}'
            ' or `custom` for your own statements.'
        )

    execution_class = cls.attach_backend(backend)
    statement_instance = execution_class(statement)  # type: BaseStatement

    return statement_instance



import logging
import warnings
from datetime import datetime
from typing import Optional, Union

from airflow.exceptions import AirflowSkipException
from airflow.models.baseoperator import BaseOperator

import deirokay
from deirokay._typing import (DeirokayDataSource, DeirokayOptionsDocument,
                              DeirokayValidationDocument)
from deirokay.enums import Backend, SeverityLevel
from deirokay.exceptions import ValidationError
from deirokay.validator import raise_validation

logger = logging.getLogger(__name__)


class DeirokayOperator(BaseOperator):
    """Parse a DataFrame from a file using Deirokay options and
    validate against a Deirokay Validation Document.
    You may choose different severity levels to mark a task as
    "soft failure" (`skipped` state) or "normal failure" (`failed`
    state).

    Parameters
    ----------
    data : Optional[str]
        File to be parsed into Deirokay.
    path_to_file : Optional[str]
        (Deprecated) File to be parsed into Deirokay.
        Use `data` instead.
    options : Union[dict, str]
        A dict or a local/S3 path to a YAML/JSON options file.
    against : Union[dict, str]
        A dict or a local/S3 path to a YAML/JSON validation
        document file.
    backend : Optional[Backend]
        Backend to use when processing data.
    template : Optional[dict]
        Map of templates to be passed to Deirokay validation.
    save_to : Optional[str], optional
        Where validation logs will be saved to.
        If None, no log is saved. By default None.
    soft_fail_level : Union[SeverityLevel, int]
        Minimum Deirokay severity level to trigger a
        "soft failure".
        Any statement with lower severity level will only raise a
        warning. Set to `None` to never trigger.
        By default SeverityLevel.MINIMAL (1).
    hard_fail_level : Union[SeverityLevel, int]
        Minimum Deirokay severity level to trigger a task failure.
        Set to `None` to never trigger.
        By default SeverityLevel.CRITICAL (5).
    reader_kwargs : Optional[dict]
        Additional keyword arguments for `Deirokay.data_reader` method.
    validator_kwargs : Optional[dict]
        Additional keyword arguments for `Deirokay.validate` method.
    **kwargs : Optional[dict]
        Additional keyword arguments for `BaseOperator`.
    """

    template_fields = [
        'data',
        'options',
        'against',
        'template',
        'save_to',
        'reader_kwargs',
        'validator_kwargs'
    ]
    template_fields_renderers = {'options': 'json', 'against': 'json'}
    ui_color = '#59f75e'

    def __init__(
        self,
        *,
        data: Optional[Union[str, DeirokayDataSource]] = None,
        # `path_to_file` is Deprecated
        path_to_file: Optional[Union[str, DeirokayDataSource]] = None,
        options: Union[str, DeirokayOptionsDocument],
        against: Union[str, DeirokayValidationDocument],
        backend: Optional[Backend] = None,
        template: Optional[dict] = None,
        save_to: Optional[str] = None,
        soft_fail_level: Union[SeverityLevel, int] = SeverityLevel.MINIMAL,
        hard_fail_level: Union[SeverityLevel, int] = SeverityLevel.CRITICAL,
        reader_kwargs: Optional[dict] = None,
        validator_kwargs: Optional[dict] = None,
        **kwargs
    ):
        super().__init__(**kwargs)

        assert bool(data) is not bool(path_to_file), (
            'Declare either `data` or `path_to_file`, but not both.'
        )
        if path_to_file:
            warnings.warn(
                'The argument `path_to_file` is deprecated and will be'
                ' removed in next major release. Use `data` instead.',
                DeprecationWarning
            )
        assert options, 'You should provide an `options` attribute.'
        assert against, 'You should provide an `against` attribute.'
        self.data = data or path_to_file
        self.options = options
        self.against = against
        self.backend = backend
        self.template = template
        self.save_to = save_to
        self.soft_fail_level = soft_fail_level
        self.hard_fail_level = hard_fail_level
        self.reader_kwargs = reader_kwargs or {}
        self.validator_kwargs = validator_kwargs or {}

    # docstr-coverage:inherited
    def execute(self, context: dict):
        current_date = datetime.strptime(context['ts_nodash'], '%Y%m%dT%H%M%S')
        df = deirokay.data_reader(
            self.data,
            options=self.options,
            backend=self.backend,
            **self.reader_kwargs
        )

        validation_document = deirokay.validate(
            df,
            against=self.against,
            template=self.template,
            save_to=self.save_to,
            current_date=current_date,
            raise_exception=False,
            **self.validator_kwargs
        )
        try:
            raise_validation(validation_document, SeverityLevel.MINIMAL)
        except ValidationError as e:
            if (
                self.hard_fail_level is not None and
                e.level >= self.hard_fail_level
            ):
                raise e
            if (
                self.soft_fail_level is not None and
                e.level >= self.soft_fail_level
            ):
                raise AirflowSkipException from e



from posixpath import splitext
from typing import Any, Dict, List, Union

import pandas  # lazy module

from deirokay.fs import fs_factory


def read(data: Union[str, 'pandas.DataFrame'],
         columns: List[str],
         sql: bool = False,
         **kwargs) -> 'pandas.DataFrame':
    """Infer the file type by its extension and call the proper
    `pandas` method to parse it.

    Parameters
    ----------
    data : Union[DataFrame, dd.DataFrame, str]
        Path to file or SQL query, or DataFrame object
    columns : List[str]
        List of columns to be parsed.
    sql : bool, optional
        Whether or not `data` should be interpreted as a path to a file
        or a SQL query.
    **kwargs : dict
        Arguments to be passed to `pandas` methods when reading.


    Returns
    -------
    pandas.DataFrame
        The pandas DataFrame.
    """
    default_kwargs: Dict[str, Any] = {}

    if isinstance(data, pandas.DataFrame):
        return data[columns].copy()

    if not isinstance(data, str):
        raise TypeError(f'Unexpected type for `data` ({data.__class__})')

    if sql:
        read_ = pandas.read_sql

    else:
        file_extension = splitext(data)[1].lstrip('.').lower()

        if file_extension == 'sql':
            query = fs_factory(data).read()
            return read(query, columns, sql=True, **kwargs)

        if file_extension in ('xls', 'xlsx'):
            file_extension = 'excel'

        if file_extension == 'csv':
            default_kwargs.update({
                'dtype': str,
                'skipinitialspace': True,
            })

        read_ = getattr(pandas, f'read_{file_extension}', None)
        if read_ is None:
            raise TypeError(f'File type "{file_extension}" not supported')

    default_kwargs.update(kwargs)

    # try `columns` argument
    try:
        return read_(data, columns=columns, **default_kwargs)
    except TypeError as e:
        if 'columns' not in str(e):
            raise e
    # try `usecols` argument
    try:
        return read_(data, usecols=columns, **default_kwargs)
    except TypeError as e:
        if 'usecols' not in str(e):
            raise e
    # give up, read everything, filter columns later
    return read_(data, **default_kwargs)[columns]



"""
Set of functions and classes related to gathering of past validation
logs from a given directory and assembling them as pandas Series.
"""

import warnings
from typing import Callable, List

import jq
import pandas  # lazy module

from ._typing import DeirokayValidationDocument
from .fs import FileSystem


def series_from_fs(series_name: str, lookback: int,
                   folder: FileSystem) -> List[DeirokayValidationDocument]:
    """List log files as FileSystem objects for a given Validation
    Document.

    Parameters
    ----------
    series_name : str
        Validation Document name.
    lookback : int
        How many logs to look behind.
    folder : FileSystem
        Where logs are read from.

    Returns
    -------
    List[FileSystem]
        List of logs to be queried.
    """

    ls = (folder/f'{series_name}/').ls(
        recursive=True, files_only=True, reverse=True, limit=lookback
    )
    return [file.read_dict() for file in ls]


class NullCallableNode():
    """Dummy node which returns nothing."""

    def __getattr__(self, name: str) -> Callable[[], None]:
        return lambda: None


class StatementNode():
    """Node at Statement level."""
    detail_keys = jq.compile('.[].report.detail | keys')

    def __init__(self, statements) -> None:
        attributes = StatementNode.detail_keys.input(statements).all()
        attributes = set([key for sub in attributes for key in sub])

        for att in attributes:
            child = (
                jq.compile(f'.[].report.detail["{att}"]')
                .input(statements).all()
            )
            setattr(self, att, pandas.Series(child))

    def __getattr__(self, name: str) -> NullCallableNode:
        return NullCallableNode()


class ItemNode():
    """Node at Item level."""
    attribute_keys = jq.compile(
        '.[].statements[] | if .alias != null then .alias else .type end'
    )

    def __init__(self, items: List[str]):
        attributes = set(ItemNode.attribute_keys.input(items).all())

        for att in attributes:
            child = jq.compile(
                '.[].statements[] | '
                f'select(.alias == "{att}" or .type == "{att}")'
            ).input(items).all()
            setattr(self, att, StatementNode(child))

    def __getattr__(self, name: str) -> StatementNode:
        return StatementNode([])


class DocumentNode():
    """Node at Document root level."""
    attribute_keys = jq.compile(
        '.[].items[] | if .alias != null then .alias else .scope end'
    )

    def __init__(self, docs: List[dict]):
        try:
            attributes = set(DocumentNode.attribute_keys.input(docs).all())
        except TypeError:
            raise TypeError(
                'All list-like scopes must be aliased'
                ' when some statement uses the `series` template.'
                ' Make sure your last Deirokay logs obey this rule.'
            )

        for att in attributes:
            child = jq.compile(
                '.[].items[] | '
                f'select(.alias == "{att}" or .scope == "{att}")'
            ).input(docs).all()
            setattr(self, att, ItemNode(child))

    def __getattr__(self, name: str):
        return ItemNode([])


def get_series(series_name: str, lookback: int,
               read_from: FileSystem) -> DocumentNode:
    """Construct a traversable object tree based on past validation
    logs. This method is aliased as `series` when called from inside
    a Validation Document.

    Parameters
    ----------
    series_name : str
        The Validation Document name.
    lookback : int
        How many logs to look behind.
    read_from : FileSystem
        Base folder where logs are read from.

    Returns
    -------
    DocumentNode
        Root node for object tree.

    Raises
    ------
    ValueError
        `read_from` can not be None.
    """
    if read_from is None:
        raise ValueError('You cannot access previous logs without providing'
                         ' a source/destination directory.')
    docs = series_from_fs(series_name, lookback, read_from)

    if not docs:
        warnings.warn('No previous log has been found in the specified folder.'
                      ' Make sure you are reading and writing to the right'
                      ' place.', Warning)

    return DocumentNode(docs)



"""
Statement to check the presence (or absence) of values in a scope.
"""
import warnings
from typing import List

import dask.dataframe  # lazy module
import numpy  # lazy module
import pandas  # lazy module

from deirokay._typing import DeirokayStatement
from deirokay._utils import noneor
from deirokay.enums import Backend
from deirokay.parser import get_dtype_treater, get_treater_instance
from deirokay.parser.loader import data_treater

from ..multibackend import profile, report
from .base_statement import BaseStatement

NODEFAULT = object()


class Contain(BaseStatement):
    """
    Checks if the given scope contains specific values. You may also
    check the number of their occurrences by specifying a minimum and
    maximum value of frequency.

    The available parameters for this statement are:

    * `rule` (required): One of `all`, `only` or `all_and_only`.
    * `values` (required): A list of values to which the rule applies.
    * `multicolumn`: A boolean indicating whether the statement should
      consider each of the `values` as a tuple of multiple columns or
      a single value. When set to False and evaluated over a scope
      containing more than one column, the rule will be applied over
      all the values from the original columns as in a single column.
      Default: False.
    * `parser`: The parser (or a list) to be used to parse the `values`.
      Correspond to the `parser` parameter of the `treater` function
      (see `deirokay.data_reader` method).
      Either `parser` or `parsers` must be declared.
    * `parsers`: An alias for `parser`, recommended when `multicolumn`
      is set to True.
      Either `parser` or `parsers` must be declared.
    * `min_occurrences`: a global minimum number of occurrences for
      each of the `values`. Default: 1 for `all` and `all_and_only`
      rules, 0 for `only`.
    * `max_occurrences`: a global maximum number of occurrences for
      each of the `values`. Default: `inf` (unbounded).
    * `occurrences_per_value`: a list of dictionaries overriding the
      global boundaries. Each dictionary may have the following keys:

        * `values` (required): a list of values to which the
          occurrence bounds below must apply to.
        * `min_occurrences`: a minimum number of occurrences for these
          values. Default: global `min_occurrences` parameter.
        * `max_occurrences`: a maximum number of occurrences for these
          values. Default: global `max_occurrences` parameter.

      Global parameters apply to all values not present in any of the
      dictionaries in `occurrences_per_value` (but yet present on the
      main `values` list).

    * `report_limit`: if set to a positive integer, limit the number of
      items generated in the statement report.
      Default: 32.

    The `all` rule checks if all the `values` declared are present in
    the scope (possibly tolerating other values not declared in
    `values`).
    Use it when you want to be sure that your data contains at least
    all the values you declare, also setting `min_occurrences` and
    `max_occurrences` when necessary.
    You may also check for "zero" occurrences of a set of values by
    setting `max_occurrences` to 0.

    The `only` rule ensures that the `values` are the only possible
    values in the scope (possibly not containing them at all).
    Use it when you want to enumerate the admitted values for the
    scope, as in an enumeration.

    The `all_and_only` rule checks both if all the `values` declared
    are present in the scope and if only they are present (not
    tolerating values not declared).
    Use it when you know all the possible values for the scope and you
    are sure that they will be always present.

    The `min_occurrences` and `max_occurrences` parameters are applied
    applied to all the `values` declared, and only these. It means you
    cannot (yet) specify boundaries for values you did't declare.

    Null values are considered valid for the purpose of the statement
    evaluation and must be explicitely passed in `values` if you wish
    to allow them (or not).

    You may also notice that, by tweaking the expected number of
    occurrences, you may end up having the very same behaviour
    regardless the `rule` you choose.
    In this case, you should go for the rule that semantically matches
    best your intents, so that your final validation document looks
    more readable and easy to understand.

    Examples
    --------
    * You have a table of users containg a column `handedness`
      only admitting the values:
      `right-handed`, `left-handed` and `ambidextrous`.
      You know that some of these values may not appear in the data,
      but you don't want other values to be present.

    .. code-block:: json

        {
            "scope": "handedness",
            "statements": [
                {
                    "type": "contain",
                    "rule": "only",
                    "values": ["right-handed", "left-handed", "ambidextrous"],
                    "parser": {"dtype": "string"}
                }
            ]
        }

    * You have a table of servers containg a column `role` which may
      contain the values `master` and `slave`.
      You want to be sure that there is always one and only one master
      server in the data.

    .. code-block:: json

        {
            "scope": "role",
            "statements": [
                {
                    "type": "contain",
                    "rule": "all",
                    "values": ["master"],
                    "parser": {"dtype": "string"},
                    "min_occurrences": 1,
                    "max_occurrences": 1
                }
            ]
        }

    You may also extend the previous example by making some adjustments
    to ensure that there is no other value than `master` and `role`
    in the data. Make notice that although the `rule` below is changed
    to `only`, the statement above is still contemplated by the
    `occurrences_per_value` parameter in the following validation item:

    .. code-block:: json

        {
            "scope": "role",
            "statements": [
                {
                    "type": "contain",
                    "rule": "only",
                    "values": ["master", "slave"],
                    "parser": {"dtype": "string"},
                    "occurrences_per_value": [
                        {
                            "values": ["master"],
                            "min_occurrences": 1,
                            "max_occurrences": 1
                        }
                    ]
                }
            ]
        }

    * You have a table of transactions containing details about
      transactions in all the branches of a company. You expect that
      there should always be at least one transaction per branch.

    .. code-block:: json

        {
            "scope": ["branch_name"],
            "statements": [
                {
                    "type": "contain",
                    "rule": "all_and_only",
                    "values": [
                        "Albany", "Utica", "Scranton", "Akron",
                        "Nashua", "Buffalo", "Rochester"
                    ],
                    "parser": {"dtype": "string"}
                }
            ]
        }

    * You have a table for the logs of user accesses to a website which
      contains an `IP` column. You want to be sure that blacklisted
      IPs are not present in the data. The following validation item in
      YAML format checks for the absense of blacklisted IPs:

    .. code-block:: yaml

        scope: IP
        statements:
        - type: contain
          rule: all
          max_occurrences: 0
          values: # blacklisted IPs
          - 3.48.48.135
          - 3.48.48.136
          parser: {dtype: string}

    * You want the pair 'San Diego' and '2022' to appear at most twice
      in your dataset (for any reason).

    .. code-block:: yaml

        scope: [city, year]
        statements:
        - type: contain
          rule: all
          min_occurrences: 0
          max_occurrences: 2
          values:
          - ['San Diego', 2022]
          parsers:
          - {dtype: string}
          - {dtype: integer}

    """
    name = 'contain'
    expected_parameters = [
        'rule',
        'values',
        'multicolumn',
        'parser',
        'parsers',
        'min_occurrences',
        'max_occurrences',
        'occurrences_per_value',
        'report_limit',
    ]
    supported_backends: List[Backend] = [Backend.PANDAS, Backend.DASK]

    DEFAULT_MIN_OCCURRENCES = {
        'all': (1, 0),
        'only': (0, 0),
        'all_and_only': (1, 0)
    }
    DEFAULT_MAX_OCCURRENCES = {
        'all': (numpy.inf, numpy.inf),
        'only': (numpy.inf, 0),
        'all_and_only': (numpy.inf, 0)
    }
    DEFAULT_REPORT_LIMIT = 32

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.rule = self.options['rule']
        assert self.rule in ('all', 'only', 'all_and_only')
        self.multicolumn = self.options.get('multicolumn', False)
        _parsers = self.options.get('parser') or self.options['parsers']
        if self.multicolumn:
            self.parsers = _parsers
        else:
            self.parsers = [_parsers]
        self.treaters = [
            get_treater_instance(parser, backend=self.get_backend())
            for parser in self.parsers
        ]

        self.min_occurrences = self.options.get('min_occurrences', None)
        self.max_occurrences = self.options.get('max_occurrences', None)
        self.occurrences_per_value = self.options.get(
            'occurrences_per_value', []
        )
        self.report_limit = self.options.get('report_limit', NODEFAULT)

        self._set_default_minmax_occurrences()

    def _set_default_minmax_occurrences(self) -> None:
        final_min = noneor(self.min_occurrences,
                           Contain.DEFAULT_MIN_OCCURRENCES[self.rule][0])
        final_max = noneor(self.max_occurrences,
                           Contain.DEFAULT_MAX_OCCURRENCES[self.rule][0])
        if (
            self.max_occurrences is not None and
            self.max_occurrences < final_min
        ):
            final_min = self.max_occurrences
        if (
            self.min_occurrences is not None and
            self.min_occurrences > final_max
        ):
            final_max = self.min_occurrences
        assert final_min >= 0
        assert final_max >= 0
        self.min_occurrences = final_min
        self.max_occurrences = final_max

    def _generate_analysis(self, value_counts):
        if self.multicolumn:
            def _unpack_row(row, *args):
                return (*row, *args)
        else:
            def _unpack_row(row, *args):
                return (row, *args)

        listed_values = [
            _unpack_row(
                row,
                self.min_occurrences,
                self.max_occurrences
            )
            for row in self.options['values']
        ]
        occurences_per_value = [
            _unpack_row(
                row,
                noneor(item.get('min_occurrences'), self.min_occurrences),
                noneor(item.get('max_occurrences'), self.max_occurrences),
            )
            for item in self.occurrences_per_value
            for row in item['values']
        ]

        occurrence_limits = pandas.DataFrame(
            occurences_per_value + listed_values,
            columns=value_counts.index.names + ['min', 'max']
        )
        options = {
            col: parser
            for col, parser in zip(value_counts.index.names, self.parsers)
        }
        data_treater(occurrence_limits, options, backend=Backend.PANDAS)

        occurrence_limits.drop_duplicates(
            subset=value_counts.index.names,
            keep='first',
            inplace=True
        )
        occurrence_limits.set_index(value_counts.index.names, inplace=True)

        analysis = value_counts.to_frame().reset_index().merge(
            occurrence_limits.reset_index(), how='outer'
        )
        analysis['count'].fillna(0, inplace=True)
        analysis['min'].fillna(Contain.DEFAULT_MIN_OCCURRENCES[self.rule][1],
                               inplace=True)
        analysis['max'].fillna(Contain.DEFAULT_MAX_OCCURRENCES[self.rule][1],
                               inplace=True)
        analysis['result'] = (
            analysis['count'].ge(analysis['min'])
            &
            analysis['count'].le(analysis['max'])
        )
        return analysis

    def _generate_report(self, analysis):
        columns = [analysis[col] for col in analysis.columns[:-4]]
        serialized = (
            treater.serialize(column)
            for treater, column in zip(self.treaters, columns)
        )
        rows = zip(*(s['values'] for s in serialized))
        values_report = sorted([
            {
                'value': value_row,
                'count': analysis_row.count,
                'result': analysis_row.result,
            }
            for value_row, analysis_row in zip(rows, analysis.itertuples())
        ], key=lambda x: x['result'])

        if (
            self.report_limit is NODEFAULT and
            len(values_report) > Contain.DEFAULT_REPORT_LIMIT
        ):
            self.report_limit = Contain.DEFAULT_REPORT_LIMIT
            warnings.warn(
                "The 'contain' statement's report size was automatically"
                f' truncated to {Contain.DEFAULT_REPORT_LIMIT} items to'
                '  prevent unexpectedly long logs.\n'
                'If you wish to set a different'
                ' size limit or even not set a limit at all (None),'
                ' please declare the `report_limit` parameter explicitely.',
                Warning
            )

        return {
            'values': (
                values_report if self.report_limit is NODEFAULT else
                values_report if self.report_limit is None else
                values_report[:self.report_limit]
            )
        }

    @report(Backend.PANDAS)
    def _report_pandas(self, df: 'pandas.DataFrame') -> dict:
        # Concat all columns
        _cols = df.columns.tolist()

        if not self.multicolumn:
            # Columns are assumed to be of same Dtype
            df = pandas.concat([df[col] for col in _cols]).to_frame()

        value_counts = (
            df.groupby(_cols, dropna=False)[_cols[0]].size()
            .rename('count')
        )
        analysis = self._generate_analysis(value_counts)
        return self._generate_report(analysis)

    @report(Backend.DASK)
    def _report_dask(self, df: 'dask.dataframe.DataFrame') -> dict:
        # Concat all columns
        _cols = df.columns.tolist()

        if not self.multicolumn:
            # Columns are assumed to be of same Dtype
            df = dask.dataframe.concat([df[col] for col in _cols]).to_frame()

        value_counts = (
            df.groupby(_cols, dropna=False)[_cols[0]].size()
            .rename('count')
        )
        analysis = self._generate_analysis(value_counts.compute())
        return self._generate_report(analysis)

    # docstr-coverage:inherited
    def result(self, report: dict) -> bool:
        return all(
            item['result'] for item in report['values']
        )

    @profile(Backend.PANDAS)
    @staticmethod
    def _profile_pandas(df: 'pandas.DataFrame') -> DeirokayStatement:
        if any(dtype != df.dtypes for dtype in df.dtypes):
            raise NotImplementedError(
                "Refusing to mix up different types of columns"
            )

        series = pandas.concat(df[col] for col in df.columns)

        unique_series = series.drop_duplicates().dropna()
        if len(unique_series) > 20:
            raise NotImplementedError("Won't generate too long statements!")

        value_frequency = series.value_counts()
        min_occurrences = int(value_frequency.min())

        statement_template = {
            'type': 'contain',
            'rule': 'all'
        }  # type: DeirokayStatement
        # Get most common type to infer treater
        try:
            statement_template.update(
                get_dtype_treater(unique_series.map(type).mode()[0])
                .attach_backend(Backend.PANDAS)
                .serialize(unique_series)  # type: ignore
            )
        except TypeError:
            raise NotImplementedError("Can't handle mixed types")
        # Sort allowing `None` values, which will appear last
        statement_template['values'].sort(key=lambda x: (x is None, x))

        if min_occurrences != 1:
            statement_template['min_occurrences'] = min_occurrences

        return statement_template

    @profile(Backend.DASK)
    @staticmethod
    def _profile_dask(df: 'dask.dataframe.DataFrame') -> DeirokayStatement:
        if any(dtype != df.dtypes for dtype in df.dtypes):
            raise NotImplementedError(
                "Refusing to mix up different types of columns"
            )

        series = dask.dataframe.concat([df[col] for col in df.columns])

        unique_series = series.drop_duplicates().dropna()
        if len(unique_series) > 20:
            raise NotImplementedError("Won't generate too long statements!")

        value_frequency = series.value_counts()
        min_occurrences = int(value_frequency.min().compute())

        statement_template = {
            'type': 'contain',
            'rule': 'all'
        }  # type: DeirokayStatement
        # Get most common type to infer treater
        try:
            statement_template.update(
                get_dtype_treater(unique_series.map(type).mode().compute()[0])
                .attach_backend(Backend.DASK)
                .serialize(unique_series)  # type: ignore
            )
        except TypeError:
            raise NotImplementedError("Can't handle mixed types")
        # Sort allowing `None` values, which will appear last
        statement_template['values'].sort(key=lambda x: (x is None, x))

        if min_occurrences != 1:
            statement_template['min_occurrences'] = min_occurrences

        return statement_template



"""
Exception types for Deirokay processes.
"""

from .enums import SeverityLevel


class ValidationError(Exception):
    """Validation failure exception."""

    def __init__(self, level: SeverityLevel, message='Validation failed'):
        self.level = level
        super().__init__(message)


class InvalidBackend(ValueError):
    """Error in backend definition."""


class UnsupportedBackend(InvalidBackend):
    """Backend is not supported for this resource."""


class ParsingError(RuntimeError):
    """Error during file parsing or object conversion."""


class InvalidStatement(SyntaxError):
    """Error when validating statement construction."""



from typing import Any, Callable, Dict, List, TypeVar, Union

import dask.dataframe  # lazy module
import pandas  # lazy module
from typing_extensions import Literal, get_args

try:
    from typing_extensions import Protocol
except ModuleNotFoundError:
    # Protocol exists only since Python 3.8
    Protocol = object

from .enums import Backend

DeirokayDataSource = TypeVar('DeirokayDataSource',
                             'pandas.DataFrame', 'dask.dataframe.DataFrame')
DeirokayDataSeries = TypeVar('DeirokayDataSeries',
                             'pandas.Series', 'dask.dataframe.Series')


# docstr-coverage:excused `No need for _typing module`
class DeirokayReadCallable(Protocol):  # noqa: E302 # docstr-coverage comment
    def __call__(self, data: Union[str, DeirokayDataSource],
                 columns: List[str],
                 *args: Any,
                 **kwargs: Any) -> DeirokayDataSource: ...

# TODO: assert if all backends are listed in TypeVars above


BackendValue = Literal['pandas', 'dask']
assert set(get_args(BackendValue)) == {backend.value for backend in Backend}

DeirokayOption = Dict[str, Any]
DeirokayColumnOptions = Dict[str, DeirokayOption]
DeirokayOptionsDocument = Dict[str, Union[str, DeirokayColumnOptions]]

DeirokayStatement = Dict[str, Any]
DeirokayValidationItem = Dict[str, Union[str, List[DeirokayStatement]]]
DeirokayValidationDocument = Dict[str,
                                  Union[str, List[DeirokayValidationItem]]]
DeirokayValidationResultDocument = Dict

DeirokaySerializedSeries = Dict[Literal['values', 'parser'], Union[List, Dict]]

AnyClass = TypeVar('AnyClass', bound=type)
AnyCallable = Callable[..., Any]
GeneralDecorator = Callable[[AnyCallable], AnyCallable]



"""
Classes and functions to treat column data types according to
Deirokay data types.
"""

from typing import Iterable, Optional

import dask.dataframe  # lazy module
import numpy  # lazy module
import pandas  # lazy module

from deirokay._typing import DeirokaySerializedSeries
from deirokay.enums import Backend, DTypes

from ..multibackend import serialize, treat
from .validator import Validator


class StringTreater(Validator):
    """Treater for string variables

    Parameters
    ----------
    treat_null_as : Optional[str], optional
        Character to replace null values for, by default None
    """
    supported_backends = [Backend.PANDAS, Backend.DASK]
    supported_dtype = DTypes.STRING
    supported_primitives = [str]

    def __init__(self, treat_null_as: Optional[str] = None, **kwargs):
        super().__init__(**kwargs)

        self.treat_null_as = treat_null_as

    @treat(Backend.PANDAS)
    def _treat_pandas(self, series: Iterable) -> 'pandas.Series':
        series = super()._treat_pandas(series)

        if self.treat_null_as is not None:
            series = series.fillna(self.treat_null_as)

        return series

    @treat(Backend.DASK)
    def _treat_dask(
        self, series: Iterable
    ) -> 'dask.dataframe.Series':
        series = super()._treat_dask(series)

        if self.treat_null_as is not None:
            series = series.fillna(self.treat_null_as)

        return series

    @staticmethod
    def _serialize_common(series):
        def _convert(item):
            if item is None or item is pandas.NA or item is numpy.NaN:
                return None
            return str(item)
        return {
            'values': [_convert(item) for item in series],
            'parser': {
                'dtype': StringTreater.supported_dtype.value
            }
        }

    @serialize(Backend.PANDAS)
    @staticmethod
    def _serialize_pandas(series: 'pandas.Series') -> DeirokaySerializedSeries:
        return StringTreater._serialize_common(series)

    @serialize(Backend.DASK)
    @staticmethod
    def _serialize_dask(series: 'dask.dataframe.Series'
                        ) -> DeirokaySerializedSeries:
        return StringTreater._serialize_common(series)



"""
Functions for data profiling and auto-generation of Deirokay Validation
Documents.
"""

import warnings
from typing import List, Optional

from deirokay.backend import detect_backend
from deirokay.enums import Backend
from deirokay.exceptions import UnsupportedBackend

from .__version__ import __version__
from ._typing import (DeirokayDataSource, DeirokayStatement,
                      DeirokayValidationDocument, DeirokayValidationItem)
from .fs import fs_factory
from .statements import STATEMENTS_MAP


def _generate_statements(df_scope: DeirokayDataSource,
                         backend: Backend) -> List[DeirokayStatement]:

    statements: List[DeirokayStatement] = []

    for stmt_cls in STATEMENTS_MAP.values():
        try:
            execution_cls = stmt_cls.attach_backend(backend)
            statement = execution_cls.profile(df_scope)
            statements.append(statement)
        except UnsupportedBackend:
            pass
        except NotImplementedError:
            pass
        except Exception as e:
            columns = list(df_scope.columns)
            warnings.warn(
                f'Unexpected error when profiling scope {columns}'
                f' using {stmt_cls.__name__} statement: {e}\n\n'
                'Please, consider reporting this issue to the '
                'developers.',
                RuntimeWarning
            )
    return statements


def _generate_items(df: DeirokayDataSource,
                    backend: Backend) -> List[DeirokayValidationItem]:
    items: List[DeirokayValidationItem] = []

    df_columns = list(df.columns)
    scope__table_stmt = [df_columns] + df_columns

    for scope in scope__table_stmt:
        df_scope = df[scope] if isinstance(scope, list) else df[[scope]]
        item = {
            'scope': scope,
            'statements': _generate_statements(df_scope, backend)
        }  # type: DeirokayValidationItem

        if item['statements']:
            items.append(item)

    return items


def profile(df: DeirokayDataSource,
            document_name: str,
            save_to: Optional[str] = None) -> DeirokayValidationDocument:
    """Generate a validation document from a given template DataFrame
    using profiling methods for builtin Deirokay statements.
    By default, statement objects are generated for the entire template
    DataFrame (the entire set of columns), and then for each of its
    columns individually.
    This function should be used only as a draft for a validation
    document or as a means to quickly launch a first version with
    minimum efforts.
    The user is encouraged to correct and supplement the generated
    document to better meet their expectations.

    Parameters
    ----------
    df : DataFrame
        The DataFrame to use as template, ideally parsed with Deirokay
        `data_reader`.
    document_name : str
        The validation document name.
    save_to : Optional[str], optional
        Path (lcaol or S3) where to save the validation document to.
        The file format is inferred by the its extension.
        If None, no document will be saved. By default None.

    Returns
    -------
    dict
        The auto-generated validation document as Python `dict`.
    """
    backend = detect_backend(df)
    validation_document = {
        'name': document_name,
        'description': f'Auto generated using Deirokay {__version__}',
        'items': _generate_items(df, backend)
    }  # type: DeirokayValidationDocument

    if save_to:
        fs_factory(save_to).write_dict(validation_document, indent=2)

    return validation_document



"""
Classes and functions to treat column data types according to
Deirokay data types.
"""

import decimal
from decimal import Decimal
from typing import Any, Iterable, List, Optional

import dask.dataframe  # lazy module
import pandas  # lazy module

from deirokay._typing import DeirokayDataSeries, DeirokaySerializedSeries
from deirokay.enums import Backend, DTypes

from ..multibackend import serialize, treat
from .float_treater import FloatTreater
from .numeric_treater import NumericTreater


class DecimalTreater(FloatTreater):
    """Treater for decimal variables

    Parameters
    ----------
    decimal_places : Optional[int], optional
        Number of decimal places, by default None
    """
    supported_backends = [Backend.PANDAS, Backend.DASK]
    supported_dtype = DTypes.DECIMAL
    supported_primitives: List[Any] = [decimal.Decimal]

    def __init__(self, decimal_places: Optional[int] = None, **kwargs):
        super().__init__(**kwargs)

        self.decimal_places = decimal_places

    def _treat_decimal_places(
        self, series: DeirokayDataSeries, **kwargs
    ) -> DeirokayDataSeries:
        if self.decimal_places is not None:
            q = Decimal(10) ** -self.decimal_places
            series = series.apply(
                lambda x: x.quantize(q) if x is not None else None,
                **kwargs
            )
        return series

    @treat(Backend.PANDAS)
    def _treat_pandas(self, series: Iterable) -> 'pandas.Series':
        series = NumericTreater._treat_pandas(self, series)
        series = self._treat_decimal_sep(series)
        series = series.map(lambda x: Decimal(x) if x is not None else None)
        series = self._treat_decimal_places(series)

        return series

    @treat(Backend.DASK)
    def _treat_dask(
        self, series: Iterable
    ) -> 'dask.dataframe.Series':
        series = NumericTreater._treat_dask(self, series)
        series = self._treat_decimal_sep(series)
        series = series.map(
            lambda x: Decimal(x) if x is not None else None,
            meta=(series.name, 'object')
        )
        series = self._treat_decimal_places(series,
                                            meta=(series.name, 'object'))

        return series

    @staticmethod
    def _serialize_common(series):
        def _convert(item):
            if item is None or item is pandas.NA:
                return None
            return str(item)
        return {
            'values': [_convert(item) for item in series],
            'parser': {
                'dtype': DecimalTreater.supported_dtype.value
            }
        }

    @serialize(Backend.PANDAS)
    @staticmethod
    def _serialize_pandas(
        series: 'pandas.Series'
    ) -> DeirokaySerializedSeries:
        return DecimalTreater._serialize_common(series)

    @serialize(Backend.DASK)
    @staticmethod
    def _serialize_dask(
        series: 'dask.dataframe.Series'
    ) -> DeirokaySerializedSeries:
        return DecimalTreater._serialize_common(series)



"""
Statement to check the number unique rows in a scope.
"""
from typing import List

import dask.dataframe  # lazy module
import pandas  # lazy module

from deirokay._typing import DeirokayStatement
from deirokay.enums import Backend

from ..multibackend import profile, report
from .base_statement import BaseStatement


class Unique(BaseStatement):
    """Checks for the unicity of rows in a scope.

    The only available option is:

    * `at_least_%`: The minimum percentage of unique rows.

    Examples
    --------
    In a table containing information about cities of your country,
    you expect the pair of columns `state` and `city` to be unique
    across all rows. It means that, although some values of `state`
    can be repeated, as well as `city` names, the combination of
    both columns should be unique.
    You can declare the following validation item to represent this
    rule:

    .. code-block:: json

        {
            "scope": ["state", "city"],
            "statements": [
                {
                    "name": "unique"
                }
            ]
        }
    """

    name = 'unique'
    expected_parameters = ['at_least_%']
    supported_backends: List[Backend] = [Backend.PANDAS, Backend.DASK]

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.at_least_perc = self.options.get('at_least_%', 100.0)

    @staticmethod
    def _unique_rows(df):
        """Get number of unique rows in DataFrame"""
        _cols = df.columns.tolist()
        value_counts = df.groupby(_cols, dropna=False)[_cols[0]].size()
        return int(sum(value_counts == 1))

    def _report_common(self, df):
        unique_rows = Unique._unique_rows(df)
        return {
            'unique_rows': unique_rows,
            'unique_rows_%': 100.0*unique_rows/len(df),
        }

    @report(Backend.PANDAS)
    def _report_pandas(self, df: 'pandas.DataFrame') -> dict:
        return self._report_common(df)

    @report(Backend.DASK)
    def _report_dask(self, df: 'dask.dataframe.DataFrame') -> dict:
        return self._report_common(df)

    # docstr-coverage:inherited
    def result(self, report: dict) -> bool:
        return report.get('unique_rows_%') >= self.at_least_perc

    @staticmethod
    def _profile_common(df):
        statement = {
            'type': 'unique',
        }  # type: DeirokayStatement

        unique_rows = Unique._unique_rows(df)
        at_least_perc = 100.0*unique_rows/len(df)

        if at_least_perc == 0.0:
            raise NotImplementedError(
                'Statement is useless when all rows are not unique.'
            )

        if at_least_perc != 100.0:
            statement['at_least_%'] = at_least_perc

        return statement

    @profile(Backend.PANDAS)
    @staticmethod
    def _profile_pandas(df: 'pandas.DataFrame') -> DeirokayStatement:
        return Unique._profile_common(df)

    @profile(Backend.DASK)
    @staticmethod
    def _profile_dask(df: 'dask.dataframe.DataFrame') -> DeirokayStatement:
        return Unique._profile_common(df)



"""
Classes and methods to build and load builtin or custom Deirokay
statements.
"""
from typing import Dict, Type

from deirokay._utils import recursive_subclass_generator

from .builtin import BaseStatement

STATEMENTS_MAP: Dict[str, Type[BaseStatement]] = {
    cls.name: cls
    for cls in recursive_subclass_generator(BaseStatement)
    if cls is not BaseStatement
}

__all__ = tuple(
    cls.__name__
    for cls in recursive_subclass_generator(BaseStatement)
)



"""
Classes and methods to load and parse correct data into Deirokay.
"""

from .loader import data_reader
from .treaters import get_dtype_treater, get_treater_instance

__all__ = (
    'data_reader',
    'get_dtype_treater',
    'get_treater_instance',
)



"""
Statement to check the number of not-null rows in a scope.
"""
import dask.dataframe  # lazy module
import pandas  # lazy module

from deirokay._typing import DeirokayStatement
from deirokay.enums import Backend

from ..multibackend import profile, report
from .base_statement import BaseStatement


class NotNull(BaseStatement):
    """Check if the rows of a scoped DataFrame are not null, possibly
    setting boundaries for the minimum and maximum percentage of
    not-null rows.

    The available options are:

    * `at_least_%`: The minimum percentage of not-null rows.
      Default: 100.0.
    * `at_most_%`: The maximum percentage of not-null rows.
      Default: 100.0.
    * `multicolumn_logic`: The logic to use when checking for not-null
      values in multicolumn scopes (either 'any' or 'all').
      Default: 'any'.

    Be careful When using multicolumn scopes: the `any` logic considers
    a row as null only if all columns are null.
    The `all` logic considers a row as null when any of its columns is
    null.

    Examples
    --------
    * You want to ensure that less than 1% of the values in a column
      `foo` are null. You can declare the following validation item:

    .. code-block:: json

        {
            "scope": "foo",
            "statements": [
                {
                    "name": "not_null",
                    "at_least_%": 99.0
                }
            ]
        }

    You noticed that you imposed a unrealistic value for `at_least_%`,
    and maybe less than 10% should be a reasonable percentage of null
    values.
    Still, you don't want to lose track of that ideal <= 1% checks,
    since you intend to improve your data quality in the near future.
    You may take advantage of `severity` to set different exception
    levels for different values of `at_least_%`:

    .. code-block:: json

        {
            "scope": "foo",
            "statements": [
                {
                    "name": "not_null",
                    "at_least_%": 99.0,
                    "severity": 3
                },
                {
                    "name": "not_null",
                    "at_least_%": 90.0,
                    "severity": 5
                }
            ]
        }

    This way, values between 90% and 99% will only raise a warning,
    while values below 90% will raise a validation exception (by
    default).

    * You don't tolerate any null values in a list of columns:

    .. code-block:: json

        {
            "scope": ["foo", "bar", "baz", "qux"],
            "statements": [
                {
                    "name": "not_null",
                    "multicolumn_logic": "all"
                }
            ]
        }

    """

    name = 'not_null'
    expected_parameters = ['at_least_%', 'at_most_%', 'multicolumn_logic']
    supported_backends = [Backend.PANDAS, Backend.DASK]

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.at_least_perc = self.options.get('at_least_%', 100.0)
        self.at_most_perc = self.options.get('at_most_%', 100.0)
        self.multicolumn_logic = self.options.get('multicolumn_logic', 'any')

        assert self.multicolumn_logic in ('any', 'all')

    def _report_common(self, df):
        if self.multicolumn_logic == 'all':
            #  REMINDER: ~all == any
            not_nulls = ~df.isnull().any(axis=1)
        else:
            not_nulls = ~df.isnull().all(axis=1)

        null_rows = int(sum(~not_nulls))
        not_null_rows = int(sum(not_nulls))
        return {
            'null_rows': null_rows,
            'null_rows_%': 100.0*null_rows/len(df),
            'not_null_rows': not_null_rows,
            'not_null_rows_%': 100.0*not_null_rows/len(df),
        }

    @report(Backend.PANDAS)
    def _report_pandas(self, df: 'pandas.DataFrame') -> dict:
        return self._report_common(df)

    @report(Backend.DASK)
    def _report_dask(self, df: 'dask.dataframe.DataFrame') -> dict:
        return self._report_common(df)

    # docstr-coverage:inherited
    def result(self, report: dict) -> bool:
        if not report.get('not_null_rows_%') >= self.at_least_perc:
            return False
        if not report.get('not_null_rows_%') <= self.at_most_perc:
            return False
        return True

    @staticmethod
    def _profile_common(df):
        statement = {
            'type': 'not_null'
        }  # type: DeirokayStatement

        not_nulls = ~df.isnull().all(axis=1)
        at_least_perc = float(100.0*sum(not_nulls)/len(not_nulls))

        if at_least_perc == 0.0:
            raise NotImplementedError(
                'Statement is useless when all rows are null.'
            )

        if at_least_perc != 100.0:
            statement['at_least_%'] = at_least_perc

        return statement

    @profile(Backend.PANDAS)
    @staticmethod
    def _profile_pandas(df: 'pandas.DataFrame') -> DeirokayStatement:
        return NotNull._profile_common(df)

    @profile(Backend.DASK)
    @staticmethod
    def _profile_dask(df: 'dask.dataframe.DataFrame') -> DeirokayStatement:
        return NotNull._profile_common(df)



"""
The base statement that all other statements inherit from.
"""
from abc import ABC, abstractmethod
from typing import List

from deirokay._typing import (DeirokayDataSource, DeirokayOption,
                              DeirokayStatement)
from deirokay.backend import MultiBackendMixin


class BaseStatement(MultiBackendMixin, ABC):
    """Base abstract statement class for all Deirokay statements.

    Attributes
    ----------
    options : dict
        Statement parameters provided by user.
    """

    name: str = 'base_statement'
    """str: Statement name when referred in Validation Documents
    (only valid for Deirokay built-in statements)."""
    expected_parameters: List[str] = ['type', 'severity', 'location']
    """List[str]: Parameters expected for this statement."""

    def __init_subclass__(cls) -> None:
        """Validate subclassed statement."""
        assert cls.name != BaseStatement.name, (
            'You should specify a `name` attribute for your statement class.'
        )

    def __init__(self, options: DeirokayOption) -> None:
        self._validate_options(options)
        self.options = options

    @classmethod
    def _validate_options(cls, options: DeirokayOption) -> None:
        """Make sure all provided statement parameters are expected
        by statement classes"""
        unexpected_parameters = [
            option for option in options
            if option not in (cls.expected_parameters +
                              BaseStatement.expected_parameters)
        ]
        if unexpected_parameters:
            raise ValueError(
                f'Invalid parameters passed to {cls.__name__} statement: '
                f'{unexpected_parameters}\n'
                f'The valid parameters are: {cls.expected_parameters}'
            )

    def __call__(self, df: DeirokayDataSource) -> dict:
        """Run statement instance."""
        internal_report = self.report(df)
        result = self.result(internal_report)

        final_report = {
            'detail': internal_report,
            'result': result
        }
        return final_report

    def report(self, df: DeirokayDataSource) -> dict:
        """Receive a DataFrame containing only columns on the scope of
        validation and returns a report of related metrics that can
        be used later to declare this Statement as fulfilled or
        failed.

        Parameters
        ----------
        df : DataFrame
            The scoped DataFrame columns to be analysed in this report
            by this statement.

        Returns
        -------
        dict
            A dictionary of useful statistics about the target columns.
        """
        raise NotImplementedError

    @abstractmethod
    def result(self, report: dict) -> bool:
        """Receive the report previously generated and declare this
        statement as either fulfilled (True) or failed (False).

        Parameters
        ----------
        report : dict
            Report generated by `report` method. Should ideally
            contain all statistics necessary to evaluate the statement
            validity.

        Returns
        -------
        bool
            Whether or not this statement passed.
        """

    @staticmethod
    def profile(df: DeirokayDataSource) -> DeirokayStatement:
        """Given a template data table, generate a statement dict
        from it.

        Parameters
        ----------
        df : DataFrame
            The DataFrame to be used as template.

        Returns
        -------
        dict
            Statement dict.

        Raises
        ------
        NotImplementedError
            If this method is not implemented by the subclass or the
            profile generation for this statement was intentionally
            skipped.
        """
        raise NotImplementedError
